#include <string.h>
#include "types.h"


static int res_cnt;
static FILE *fp;
static UINT32 file_size;


static struct {
	UINT16 index_num;
	UINT32 data_ptr;
} midi_map[65536];


extern void Dynamix_LZW_decode( UINT32, UINT8 * );


void Dump_File( int decoder, FILE *fp_in, UINT32 in_size, UINT32 out_size, int file_no, char *ext )
{
	FILE *fp_out;
	char str[256];

	//========================================
	//========================================
	//========================================

	// raw format
	if( decoder == 0 ) {
		fread( &data_out, 1, in_size, fp_in );
	}

	else if( decoder == 2 ) {
		reset_file( data_in, in_size );
		fread( &data_in, 1, in_size, fp_in );


		Dynamix_LZW_decode( out_size, data_out );
	}

	//========================================
	//========================================
	//========================================

	sprintf( str, "%d.%s", file_no, ext );
	fp_out = fopen( str, "wb" );
	if( !fp_out ) return;


	fwrite( &data_out, 1, out_size, fp_out );
	fclose(fp_out);
}


void Dump_File_Str( int decoder, FILE *fp_in, UINT32 in_size, UINT32 out_size, char *name, char *ext )
{
	FILE *fp_out;
	char str[256];

	//========================================
	//========================================
	//========================================

	// raw format
	if( decoder == 0 ) {
		fread( &data_out, 1, in_size, fp_in );
	}

	else if( decoder == 2 ) {
		reset_file( data_in, in_size );
		fread( &data_in, 1, in_size, fp_in );


		Dynamix_LZW_decode( out_size, data_out );
	}

	//========================================
	//========================================
	//========================================

	sprintf( str, "%s.%s", name, ext );
	fp_out = fopen( str, "wb" );
	if( !fp_out ) return;


	fwrite( &data_out, 1, out_size, fp_out );
	fclose(fp_out);
}


void Dynamix_Scan_SND()
{
	UINT32 in_ptr;
	UINT32 map_cnt;
	UINT16 map_total;
	char str_id[4+1];


	in_ptr = 8;
	map_cnt = 0;
	map_total = 65535;
	str_id[4] = 0;


	while( map_cnt < map_total ) {
		fseek( fp, in_ptr, SEEK_SET );
		fread( &str_id, 1, 4, fp );


		if( strcmp( str_id, "INF:" ) == 0 ) {
			UINT32 map_size;
			UINT16 dummy1;
			UINT8 dummy2;


			/*
			header:
			4-byte packed size (after entry)

			2-byte ????
			2-byte # entries
			1-byte ????
			*/

			fread( &map_size, 1, 4, fp );
			fread( &dummy1, 1, 2, fp );
			fread( &map_total, 1, 2, fp );
			fread( &dummy2, 1, 1, fp );


			/*
			table:
			2-byte index #
			4-byte absolute ptr (after this entry)
			*/

			for( int lcv=0; lcv < map_total; lcv++ ) {
				fread( &midi_map[lcv].index_num, 1, 2, fp );
				fread( &midi_map[lcv].data_ptr, 1, 4, fp );
			}
		}

		else if( strcmp( str_id, "DAT:" ) == 0 ) {
			UINT32 in_size, out_size;
			UINT16 index_num, dummy1;
			UINT8 compression;

			/*
			4-byte packed size (after this entry)

			2-byte index #
			2-byte ??? resource type (0009 = midi)
			1-byte ??? compression (00 = raw)
			4-byte output size (after this entry)
			*/

			fread( &in_size, 1, 4, fp );
			fread( &index_num, 1, 2, fp );
			fread( &dummy1, 1, 2, fp );
			fread( &compression, 1, 1, fp );
			fread( &out_size, 1, 4, fp );


			in_size -= (2 + 2 + 1 + 4);
			Dump_File( compression, fp, in_size, out_size, index_num, "snd" );


			printf( "%d.snd | Offset = %X\n",
				index_num,
				in_ptr );


			map_cnt++;
		}


		in_ptr = midi_map[map_cnt].data_ptr;
	}
}


void Dynamix_Scan_SSM( UINT32 stop )
{
	UINT32 in_ptr;
	char str_id[4+1];


	in_ptr = 8;
	str_id[4] = 0;


	while( in_ptr < stop ) {
		UINT32 in_size, out_size;
		UINT8 compression;


		/*
		4-byte ID
		4-byte packed size (after this entry)
		1-byte compression (00 = raw)
		4-byte output size (after this entry)
		*/

		fseek( fp, in_ptr, SEEK_SET );
		fread( &str_id, 1, 4, fp );
		fread( &in_size, 1, 4, fp );
		fread( &compression, 1, 1, fp );
		fread( &out_size, 1, 4, fp );


		str_id[3] = 0;


		Dump_File_Str( compression, fp, in_size-5, out_size, str_id, "drv" );


		printf( "%s.DRV | Offset = %X\n",
			str_id,
			in_ptr );


		in_ptr += (4 + 4 + in_size);
	}
}


void Dump_Dynamix( int argc, char **argv )
{
	char str[4+1];
	UINT8 dummy;


	if( argc < 2+1 ) return;

	str[4] = 0;


	/*
	4-byte ID
	3-byte size (after full header)
	1-byte flag? ($80)
	*/

	fp = fopen( argv[2], "rb" );
	if( !fp ) return;


	fread( &str, 1, 4, fp );
	fread( &file_size, 1, 3, fp );
	fread( &dummy, 1, 1, fp );


	// add header
	file_size += 8;


	if( strcmp( str, "SND:" ) == 0 )
		Dynamix_Scan_SND();
	else if( strcmp( str, "SSM:" ) == 0 )
		Dynamix_Scan_SSM( file_size );
}
