#include <string.h>
#include "types.h"


/*
80 = View
81 = Picture
82 = Script (0082)
83 = Text (0083)
84 = Sound
85 = Etc
86 = Vocab (0086)
87 = Font
88 = Cursor
89 = Patch
*/

char *sci0_type[] = {
	"view",
	"pic",
	"script",
	"text",
	"sound",
	"etc",
	"vocab",
	"font",
	"cursor",
	"patch",
};



/*
80 = View
81 = Picture
82 = Script (0082)
83 = Text (0083)
84 = Sound (0084)
85 = Etc
86 = Vocab (0086)
87 = Font
88 = Cursor
89 = Patch
8A = Bitmap
8B = Palette
8C = Wav
8D = Audio (PCM)
8E = Sync
8F = Message
90 = Map
91 = Heap
92 = Audio36
93 = Sync36
94 = Xlate
*/

char *sci1_type[] = {
	"v56",
	"p56",
	"scr",
	"tex",
	"snd",
	"etc",
	"voc",
	"fon",
	"cur",
	"pat",
	"bmp",
	"pal",
	"wav",
	"aud",
	"syn",
	"msg",

	"map",
	"hep",
	"a36",
	"s36",
	"trn",
};



static struct {
	UINT8 res_type;
	UINT16 res_num;

	UINT8 res_file;
	UINT32 res_offset;
} sci_resmap[65536];


static int res_cnt;



extern void Sierra_LZW_AGI_decode( UINT32, UINT8 * );
extern void Sierra_LZW_SCI1_decode( UINT32, UINT8 * );
extern void Sierra_LZS_decode( UINT32, UINT8 * );
extern void Sierra_DCL_unpack(void *p,void *u);



void Dump_Raw_Resource( char *name, int index, char *type )
{
	FILE *fp, *fp_out;
	char str[256];

	UINT32 res_packed, res_size;



	sprintf( str, "%s", name );
	fp = fopen( str, "rb" );
	if( !fp ) return;



	fseek( fp, 0, SEEK_END );
	res_packed = ftell(fp);
	fseek( fp, 0, SEEK_SET );
	
	//========================================
	//========================================
	//========================================

	fread( &data_out, 1, res_packed, fp );
	
	res_size = res_packed;

	//========================================
	//========================================
	//========================================

	sprintf( str, "%s.%03d",
		type, index );


	fp_out = fopen( str, "wb" );
	if( fp_out ) {
		// header already included!!


		fwrite( &data_out, 1, res_size, fp_out );
		fclose(fp_out);
	}



	fclose(fp);
}



void Dump_AGI_Resource( char *name, int index, char *type )
{
	FILE *fp, *fp_out;
	char str[256];

	UINT32 res_packed, res_size;



	sprintf( str, "%s", name );
	fp = fopen( str, "rb" );
	if( !fp ) return;



	fseek( fp, 0, SEEK_END );
	res_packed = ftell(fp);
	fseek( fp, 0, SEEK_SET );
	
	//========================================
	//========================================
	//========================================

	/*
	4-byte header
	
	1 = decoded size
	*/

	fread( &res_size, 1, 4, fp );


	reset_file( data_in, res_packed-4 );
	fread( &data_in, 1, res_packed-4, fp );


	Sierra_LZW_AGI_decode( res_size, data_out );

	//========================================
	//========================================
	//========================================

	sprintf( str, "%s.%03d",
		type, index );


	fp_out = fopen( str, "wb" );
	if( fp_out ) {
		// header already included!!


		fwrite( &data_out, 1, res_size, fp_out );
		fclose(fp_out);
	}



	fclose(fp);
}



void Dump_SCI0_Resource( int index )
{
	FILE *fp, *fp_out;
	char str[256];

	UINT16 res_type, res_packed, res_size, res_decoder;



	sprintf( str, "RESOURCE.%03d", sci_resmap[index].res_file );
	fp = fopen( str, "rb" );
	if( !fp ) return;


	fseek( fp, sci_resmap[index].res_offset, SEEK_SET );
	
	//========================================
	//========================================
	//========================================

	/*
	4-word (16-bit) header
	
	1 = type / num
	2 = packed size (w/ 4 header bytes)
	3 = decoded size
	4 = compression
	*/

	fread( &res_type, 1, 2, fp );
	fread( &res_packed, 1, 2, fp );
	fread( &res_size, 1, 2, fp );
	fread( &res_decoder, 1, 2, fp );



	// raw format
	if( res_decoder == 0 ) {
		fread( &data_in, 1, res_size, fp );
	}

	else if( res_decoder == 1 ) {
		reset_file( data_in, res_packed-4 );
		fread( &data_in, 1, res_packed-4, fp );


		Sierra_LZW_AGI_decode( res_size, data_out );
	}

	//========================================
	//========================================
	//========================================

	sprintf( str, "%s.%03d",
		sci0_type[ sci_resmap[index].res_type ],
		sci_resmap[index].res_num );


	fp_out = fopen( str, "wb" );
	if( fp_out ) {
		UINT16 id;


		// print 16-bit ID header
		id = sci_resmap[index].res_type | 0x80;


		fwrite( &id, 1, 2, fp_out );
		fwrite( &data_out, 1, res_size, fp_out );
		fclose(fp_out);
	}



	fclose(fp);
}



void Dump_SCI1_Resource( int type, int index )
{
	FILE *fp, *fp_out;
	char str[256];

	UINT8 res_type;
	UINT16 res_id, res_packed, res_size, res_decoder;



	sprintf( str, "RESOURCE.%03d", sci_resmap[index].res_file );
	fp = fopen( str, "rb" );
	if( !fp ) return;


	fseek( fp, sci_resmap[index].res_offset, SEEK_SET );

	//========================================
	//========================================
	//========================================

	if( type == 1 ) {
		/*
		4-word (16-bit) header
		
		1 = type / num
		2 = packed size (w/ 4 header bytes)
		3 = decoded size
		4 = compression
		*/

		fread( &res_id, 1, 1, fp );
		fread( &res_type, 1, 1, fp );
		fread( &res_packed, 1, 2, fp );
		fread( &res_size, 1, 2, fp );
		fread( &res_decoder, 1, 2, fp );
	}
	

	else if( type == 2 ) {
		/*
		1-byte ID header
		1 = type


		4-word (16-bit) header
		
		2 = num
		3 = packed size (w/ 4 header bytes)
		4 = decoded size
		5 = compression
		*/

		fread( &res_type, 1, 1, fp );
		fread( &res_id, 1, 2, fp );
		fread( &res_packed, 1, 2, fp );
		fread( &res_size, 1, 2, fp );
		fread( &res_decoder, 1, 2, fp );
	}



	// raw format
	if( res_decoder == 0 ) {
		fread( &data_out, 1, res_size, fp );
	}

	else if( res_decoder == 2 ) {
		reset_file( data_in, res_packed-4 );
		fread( &data_in, 1, res_packed-4, fp );


		Sierra_LZW_SCI1_decode( res_size, data_out );
	}

	else if( res_decoder >= 18 && res_decoder <= 20 ) {
		reset_file( data_in, res_packed );
		fread( &data_in, 1, res_packed, fp );


		Sierra_DCL_unpack( (void *) data_in, (void *) data_out );
	}

	//========================================
	//========================================
	//========================================

	sprintf( str, "%d.%s",
		sci_resmap[index].res_num,
		sci1_type[ sci_resmap[index].res_type ] );


	fp_out = fopen( str, "wb" );
	if( fp_out ) {
		UINT16 id;


		// print 16-bit ID header
		id = sci_resmap[index].res_type | 0x80;


		fwrite( &id, 1, 2, fp_out );
		fwrite( &data_out, 1, res_size, fp_out );
		fclose(fp_out);
	}



	fclose(fp);
}


void Dump_SCI2_Resource( int index )
{
	FILE *fp, *fp_out;
	char str[256];

	UINT8 res_type;
	UINT16 res_id, res_decoder;
	UINT32 res_packed, res_size;



	sprintf( str, "RESOURCE.%03d", sci_resmap[index].res_file );
	fp = fopen( str, "rb" );
	if( !fp ) return;


	fseek( fp, sci_resmap[index].res_offset, SEEK_SET );
	
	//========================================
	//========================================
	//========================================

	/*
	1-byte ID header
	1 = type


	2 = num (16-bit)
	3 = packed size (32-bit)
	4 = decoded size (32-bit)
	5 = compression (16-bit)
	*/

	fread( &res_type, 1, 1, fp );
	fread( &res_id, 1, 2, fp );
	fread( &res_packed, 1, 4, fp );
	fread( &res_size, 1, 4, fp );
	fread( &res_decoder, 1, 2, fp );



	// raw format
	if( res_decoder == 0 ) {
		fread( &data_out, 1, res_size, fp );
	}

	else if( res_decoder == 32 ) {
		reset_file( data_in, res_packed );
		fread( &data_in, 1, res_packed, fp );


		Sierra_LZS_decode( res_size, data_out );
	}

	//========================================
	//========================================
	//========================================

	sprintf( str, "%d.%s",
		sci_resmap[index].res_num,
		sci1_type[ sci_resmap[index].res_type ] );


	fp_out = fopen( str, "wb" );
	if( fp_out ) {
		UINT16 id;


		// print 16-bit ID header
		id = sci_resmap[index].res_type | 0x80;


		fwrite( &id, 1, 2, fp_out );
		fwrite( &data_out, 1, res_size, fp_out );
		fclose(fp_out);
	}



	fclose(fp);
}


void Scan_SCI0( int sci_mode, char *type )
{
	FILE *fp;
	int sci_target;


	sci_target = -1;


	if( strcmp( type, "all" ) == 0 ) sci_target = 255;
	else {
		for( int lcv=0; lcv<10; lcv++ ) {
			if( strcmp( type, sci0_type[lcv] ) == 0 ) { 
				sci_target = lcv;
				break;
			}
		}
	}

	
	fp = fopen( "resource.map", "rb" );
	if( !fp ) return;



	while( !feof(fp) ) {
		UINT16 buf1;
		UINT32 buf2;


		// 6-byte packs
		fread( &buf1, 1, 2, fp );
		fread( &buf2, 1, 4, fp );
		if( buf1 == (UINT16) 0xffff && buf2 == (UINT32) 0xffffffff )
			break;


		// SCI0 - type 1
		if( sci_mode == 1 ) {
			sci_resmap[res_cnt].res_type = buf1 >> 11;
			sci_resmap[res_cnt].res_num = buf1 & 0x7ff;

			sci_resmap[res_cnt].res_file = (UINT8) (buf2 >> 26);
			sci_resmap[res_cnt].res_offset = buf2 & 0x03ffffff;
		}




		printf( "%s.%03d | File = resource.%03d   Offset = %X\n",
			sci0_type[ sci_resmap[res_cnt].res_type ],
			sci_resmap[res_cnt].res_num,
			sci_resmap[res_cnt].res_file,
			sci_resmap[res_cnt].res_offset );


		if( sci_target == 255 || sci_resmap[res_cnt].res_type == sci_target )
			Dump_SCI0_Resource( res_cnt );


		res_cnt++;
	}


	printf( "total = %d\n", res_cnt );
}


void Scan_SCI1_1( int sci_mode, char *type )
{
	FILE *fp;
	int sci_target;


	sci_target = -1;


	if( strcmp( type, "all" ) == 0 ) sci_target = 255;
	else {
		for( int lcv=0; lcv<10; lcv++ ) {
			if( strcmp( type, sci0_type[lcv] ) == 0 ) { 
				sci_target = lcv;
				break;
			}
		}
	}


	fp = fopen( "resource.map", "rb" );
	if( !fp ) return;



	while( !feof(fp) ) {
		UINT16 buf1;
		UINT32 buf2;


		// 6-byte packs
		fread( &buf1, 1, 2, fp );
		fread( &buf2, 1, 4, fp );
		if( buf1 == (UINT16) 0xffff && buf2 == (UINT32) 0xffffffff )
			break;


		// SCI1 - type 1
		if( sci_mode == 1 ) {
			sci_resmap[res_cnt].res_type = buf1 >> 11;
			sci_resmap[res_cnt].res_num = buf1 & 0x7ff;

			sci_resmap[res_cnt].res_file = (UINT8) (buf2 >> 26);
			sci_resmap[res_cnt].res_offset = buf2 & 0x03ffffff;
			//sci_resmap[res_cnt].res_file = (UINT8) (buf2 >> 28);
			//sci_resmap[res_cnt].res_offset = buf2 & 0x0fffffff;
		}



		printf( "%s.%03d | File = resource.%03d   Offset = %X\n",
			sci0_type[ sci_resmap[res_cnt].res_type ],
			sci_resmap[res_cnt].res_num,
			sci_resmap[res_cnt].res_file,
			sci_resmap[res_cnt].res_offset );


		if( sci_target == 255 || sci_resmap[res_cnt].res_type == sci_target )
			Dump_SCI1_Resource( 1, res_cnt );


		res_cnt++;
	}


	printf( "total = %d\n", res_cnt );
}


void Scan_SCI1_2( int sci_mode, char *type )
{
	FILE *fp;
	int sci_target;
	int lcv;
	
	UINT8 res_type[16];
	UINT16 res_ofs[16];



	sci_target = -1;


	if( strcmp( type, "all" ) == 0 ) sci_target = 255;
	else {
		for( int lcv=0; lcv<21; lcv++ ) {
			if( strcmp( type, sci1_type[lcv] ) == 0 ) { 
				sci_target = lcv;
				break;
			}
		}
	}

	
	fp = fopen( "resource.map", "rb" );
	if( !fp ) return;



	res_cnt = 0;
	while( !feof(fp) ) {
		UINT8 buf1;
		UINT16 buf2;


		// header table
		fread( &buf1, 1, 1, fp );
		fread( &buf2, 1, 2, fp );

		res_type[ res_cnt ] = buf1;
		res_ofs[ res_cnt ] = buf2;


		if( buf1 == 0xff ) break;
		res_cnt++;
	}



	fseek( fp, res_ofs[0], SEEK_SET );


	lcv = 0;
	res_cnt = 0;
	while( res_type[lcv] != 0xff ) {
		// SCI1 - type 2
		if( sci_mode == 2 ) {
			UINT16 buf1;
			UINT32 buf2;

			// 6-byte packs
			fread( &buf1, 1, 2, fp );
			fread( &buf2, 1, 4, fp );



			sci_resmap[res_cnt].res_type = res_type[lcv] & 0x7f;
			sci_resmap[res_cnt].res_num = buf1;

			sci_resmap[res_cnt].res_file = (UINT8) (buf2 >> 28);
			sci_resmap[res_cnt].res_offset = buf2 & 0x0fffffff;
		}

		// SCI1 - type 3
		if( sci_mode == 3 ) {
			UINT16 buf1;
			UINT32 buf2;


			// 5-byte packs
			fread( &buf1, 1, 2, fp );
			fread( &buf2, 1, 3, fp );



			sci_resmap[res_cnt].res_type = res_type[lcv] & 0x7f;
			sci_resmap[res_cnt].res_num = buf1;

			sci_resmap[res_cnt].res_file = 0;
			sci_resmap[res_cnt].res_offset = (buf2 & 0xffffff) * 2;
		}



		printf( "%s.%03d | File = resource.%03d   Offset = %X\n",
			sci1_type[ sci_resmap[res_cnt].res_type ],
			sci_resmap[res_cnt].res_num,
			sci_resmap[res_cnt].res_file,
			sci_resmap[res_cnt].res_offset );


		if( sci_target == 255 || sci_resmap[res_cnt].res_type == sci_target )
			Dump_SCI1_Resource( 2, res_cnt );


		if( ftell(fp) >= res_ofs[lcv+1] ) lcv++;


		res_cnt++;
	}

	
	printf( "total = %d\n", res_cnt );
}


void Scan_SCI2( int sci_mode, char *type )
{
	FILE *fp;
	int sci_target;
	int lcv;
	
	UINT8 res_type[16];
	UINT16 res_ofs[16];



	sci_target = -1;


	if( strcmp( type, "all" ) == 0 ) sci_target = 255;
	else {
		for( int lcv=0; lcv<21; lcv++ ) {
			if( strcmp( type, sci1_type[lcv] ) == 0 ) { 
				sci_target = lcv;
				break;
			}
		}
	}


	fp = fopen( "resource.map", "rb" );
	if( !fp ) return;



	res_cnt = 0;
	while( !feof(fp) ) {
		// SCI2 - type 1
		if( sci_mode == 1 ) {
			UINT8 buf1;
			UINT16 buf2;


			fread( &buf1, 1, 1, fp );
			fread( &buf2, 1, 2, fp );

			res_type[ res_cnt ] = buf1;
			res_ofs[ res_cnt ] = buf2;


			if( buf1 == 0xff ) break;
		}

		// SCI2 - type 2
		else if( sci_mode == 2 ) {
			// 16-bit ID
			// 32-bit offset
		}

		res_cnt++;
	}


	fseek( fp, res_ofs[0], SEEK_SET );


	lcv = 0;
	res_cnt = 0;
	while( res_type[lcv] != 0xff ) {
		// SCI2 - type 1
		if( sci_mode == 0 ) {
			UINT16 buf1;
			UINT32 buf2;


			// 6-byte packs
			fread( &buf1, 1, 2, fp );
			fread( &buf2, 1, 4, fp );



			sci_resmap[res_cnt].res_type = res_type[lcv] & 0x7f;
			sci_resmap[res_cnt].res_num = buf1;

			sci_resmap[res_cnt].res_file = 0;
			sci_resmap[res_cnt].res_offset = buf2 & 0xffffffff;
		}



		printf( "%s.%03d | File = resource.%03d   Offset = %X\n",
			sci1_type[ sci_resmap[res_cnt].res_type ],
			sci_resmap[res_cnt].res_num,
			sci_resmap[res_cnt].res_file,
			sci_resmap[res_cnt].res_offset );


		if( sci_target == 255 || sci_resmap[res_cnt].res_type == sci_target )
			Dump_SCI2_Resource( res_cnt );


		if( ftell(fp) >= res_ofs[lcv+1] ) lcv++;


		res_cnt++;
	}

	
	printf( "total = %d\n", res_cnt );
}


void Scan_Misc( int sci_mode, char *file, char *ext )
{
	/*
	Silpheed
	- RESOURCE.0xx (raw)
	*/

	if( sci_mode == 1 ) {
		for( int lcv=0; lcv<999; lcv++ ) {
			char str[256];

			sprintf( str, "%s%03d", file, lcv );


			Dump_Raw_Resource( str, lcv, ext );
		}
	}



	/*
	Sorcerian
	- SS0xx (LZW)
	- SCENE03\SS099 != SONG\SS099
	- SCENE03\SS059 != SCENE01\SS059
	*/

	else if( sci_mode == 2 ) {
		for( int lcv=0; lcv<999; lcv++ ) {
			char str[256];

			sprintf( str, "%s%03d", file, lcv );


			Dump_AGI_Resource( str, lcv, ext );
		}
	}
}


void Dump_Sierra( int argc, char **argv )
{
	if( argc < 3+1 ) return;



	if( strcmp( argv[2], "misc_1" ) == 0 )
		Scan_Misc( 1, argv[3], argv[4] );

	else if( strcmp( argv[2], "misc_2" ) == 0 )
		Scan_Misc( 2, argv[3], argv[4] );

	// -----------------------------------------------
	// -----------------------------------------------

	else if( strcmp( argv[2], "sci0" ) == 0 )
		Scan_SCI0( 1, argv[3] );

	// -----------------------------------------------
	// -----------------------------------------------

	else if( strcmp( argv[2], "sci1_1" ) == 0 )
		Scan_SCI1_1( 1, argv[3] );

	else if( strcmp( argv[2], "sci1_2" ) == 0 )
		Scan_SCI1_2( 2, argv[3] );

	else if( strcmp( argv[2], "sci1_3" ) == 0 )
		Scan_SCI1_2( 3, argv[3] );

	// -----------------------------------------------
	// -----------------------------------------------

	else if( strcmp( argv[2], "sci2" ) == 0 )
		Scan_SCI2( 1, argv[3] );
}
