/*
SCI2


LZS (LZ77 variant)



Thanks to: VAG's SCI Tools
*/

#include <memory.h>
#include "types.h"



void Sierra_LZS_reset()
{
}


void Sierra_LZS_decode( UINT32 len, UINT8 *buf_out )
{
	UINT32 out_ptr;


	Sierra_LZS_reset();


	out_ptr = 0;
	while( len > 0 )
	{
		UINT32 code;
		UINT32 lz_ofs, lz_len, lz_ptr;


		// get next code
		code = get_bits_left(1);
		if( code == 0xffffffff ) return;


		// raw = 0 | xxxxxxxx
		if( code == 0 ) {
			buf_out[ out_ptr++ ] = (UINT8) get_bits_left(8);
			len--;
			continue;
		}



		// LZ offset - 7 or 11 bits
		if( get_bits_left(1) == 1 ) {
			lz_ofs = get_bits_left(7);


			// Abort code!!
			if( lz_ofs == 0 ) break;
		}
		else
			lz_ofs = get_bits_left(11);



		/*
		2 = 00
		3 = 01

		4 = 10

		5 = 1100
		6 = 1101

		7 = 1110


		(repeat 1111)
		8-22 = 1111 xxxx
		23-37 = 1111 1111 xxxx
		38+ = ???


		end = 1100 00000
		*/

		// LZ length - 2-12+ bits

		// x
 		if( get_bits_left(1) == 0 )
			// 0(x)
			lz_len = 2 + get_bits_left(1);

		// 1x
		else if( get_bits_left(1) == 0 )
			// 10
			lz_len = 4;

		// 11x
		else if( get_bits_left(1) == 0 )
			// 110(x)
			lz_len = 5 + get_bits_left(1);

		// 111x
		else if( get_bits_left(1) == 0 )
			// 1110
			lz_len = 7;

		// 1111 xxxx
		else {
			lz_len = 8;

			// 4-bit packs
			while(1) {
				code = get_bits_left(4);
				if( code != 15 ) break;

				lz_len += 15;
			}

			lz_len += code;
		}



		// creates errors!!
		//memcpy( buf_out + out_ptr, buf_out - lz_ofs, lz_len );


		// do LZ copy
		lz_ptr = out_ptr - lz_ofs;
		while( lz_len > 0 ) {
			buf_out[ out_ptr++ ] = buf_out[ lz_ptr++ ];

			lz_len--;
			len--;
		}
	} // end len
}
