/*
AGI - SCI0


Adaptive LZW (expands dictionary)



Thanks to: FreeSCI + ScummVM
*/

#include <memory.h>
#include "types.h"



static UINT8 code_cur[256];
static UINT32 code_size;
static UINT32 code_len;


static struct {
	UINT8 str[256];
	UINT8 len;
} dict_table[0x4000];

static UINT32 dict_size;
static UINT32 dict_max;
static bool dict_full;



void Sierra_LZW_AGI_reset()
{
	memset( &dict_table, 0, sizeof(dict_table) );

	for( int lcv=0; lcv<256; lcv++ ) {
		dict_table[lcv].len = 1;
		dict_table[lcv].str[0] = lcv;
	}


	// 00-FF = ASCII
	// 100 = reset
	// 101 = stop
	dict_size = 0x102;
	dict_max = 0x200;
	dict_full = false;


	// start = 9 bit codes
	code_size = 9;
	code_len = 0;
}


void Sierra_LZW_AGI_decode( UINT32 len, UINT8 *buf_out )
{
	UINT32 out_ptr;


	Sierra_LZW_AGI_reset();


	out_ptr = 0;
	while( len >= 0 )
	{
		UINT32 code;
		UINT32 lcv;
		bool hit;


		// get next code
		code = get_bits_right( code_size );
		if( code == 0xffffffff ) return;


		// reset
		if( code == 0x100 ) {
			Sierra_LZW_AGI_reset();
			continue;
		}


		// stop
		if( code == 0x101 ) {
			break;
		}



		// special case: expand for new entry
		if( code >= dict_size && dict_full == false ) {
			code_cur[ code_len++ ] = code_cur[0];


			// write output - future expanded string
			for( lcv=0; lcv<code_len; lcv++ )
				buf_out[ out_ptr++ ] = code_cur[lcv];
		}
		else {
			// write output
			for( lcv=0; lcv<dict_table[code].len; lcv++ )
				buf_out[ out_ptr++ ] = dict_table[code].str[lcv];


			// expand string
			code_cur[ code_len++ ] = dict_table[code].str[0];
		}



		// add to dictionary: 2+ bytes only
		if( code_len >= 2 ) hit = false;
		else hit = true;



		// add to dictionary
		if( hit == false ) {
			if( dict_full == false ) {
				// check full condition
				if( dict_size == dict_max && code_size == 12 ) {
					dict_full = true;

					lcv = dict_size;
				}
				else
					lcv = dict_size++;


				// expand dictionary (adaptive LZW)
				if( dict_size == dict_max && code_size < 12 ) {
					dict_max *= 2;
					code_size++;
				}


				// add new entry
				for( UINT32 lcv2=0; lcv2<code_len; lcv2++ ) {
					dict_table[lcv].str[lcv2] = code_cur[lcv2];
					dict_table[lcv].len++;
				}
			}


			// reset running code!
			for( lcv=0; lcv<dict_table[code].len; lcv++ )
				code_cur[lcv] = dict_table[code].str[lcv];

			code_len = dict_table[code].len;
		}
	} // end len
}
