// Dynamix RMF info: ScummVM DGDS, NewRisingSun


#include <string.h>
#include "types.h"


static int res_cnt;
static FILE *fp;
static UINT32 file_size;


static struct {
	UINT16 index_num;
	UINT32 data_ptr;
	char name[80];
} sound_map[65536];


static struct {
	UINT8 file_num;


	// table
	UINT32 hash;
	UINT32 offset;
} rmf_map[65536];

UINT8 rmf_names[65536][13];




extern void Dynamix_RLE_decode( UINT32, UINT8 * );
extern void Dynamix_LZW_decode( UINT32, UINT8 * );



void Dump_File( int decoder, FILE *fp_in, UINT32 in_size, UINT32 out_size, int file_no, char *ext )
{
	FILE *fp_out;
	char str[256];

	//========================================
	//========================================
	//========================================

	// raw format
	if( decoder == 0 ) {
		fread( &data_out, 1, in_size, fp_in );
	}

	else if( decoder == 1 ) {
		reset_file( data_in, in_size );
		fread( &data_in, 1, in_size, fp_in );


		Dynamix_RLE_decode( out_size, data_out );
	}

	else if( decoder == 2 ) {
		reset_file( data_in, in_size );
		fread( &data_in, 1, in_size, fp_in );


		Dynamix_LZW_decode( out_size, data_out );
	}

	//========================================
	//========================================
	//========================================

	sprintf( str, "%d.%s", file_no, ext );
	fp_out = fopen( str, "wb" );
	if( !fp_out ) return;


	fwrite( &data_out, 1, out_size, fp_out );
	fclose(fp_out);
}



void Dump_File_Ext( int decoder, FILE *fp_in, UINT32 in_size, UINT32 out_size, char *name, char *ext )
{
	FILE *fp_out;
	char str[256];

	//========================================
	//========================================
	//========================================

	// raw format
	if( decoder == 0 ) {
		fread( &data_out, 1, in_size, fp_in );
	}

	else if( decoder == 1 ) {
		reset_file( data_in, in_size );
		fread( &data_in, 1, in_size, fp_in );


		Dynamix_RLE_decode( out_size, data_out );
	}

	else if( decoder == 2 ) {
		reset_file( data_in, in_size );
		fread( &data_in, 1, in_size, fp_in );


		Dynamix_LZW_decode( out_size, data_out );
	}

	//========================================
	//========================================
	//========================================

	sprintf( str, "%s.%s", name, ext );
	fp_out = fopen( str, "wb" );
	if( !fp_out ) return;


	fwrite( &data_out, 1, out_size, fp_out );
	fclose(fp_out);
}



void Dump_File_Name( int decoder, FILE *fp_in, UINT32 in_size, UINT32 out_size, char *name )
{
	FILE *fp_out;
	char str[256];

	//========================================
	//========================================
	//========================================

	// raw format
	if( decoder == 0 ) {
		fread( &data_out, 1, in_size, fp_in );
	}

	else if( decoder == 1 ) {
		reset_file( data_in, in_size );
		fread( &data_in, 1, in_size, fp_in );


		Dynamix_RLE_decode( out_size, data_out );
	}

	else if( decoder == 2 ) {
		reset_file( data_in, in_size );
		fread( &data_in, 1, in_size, fp_in );


		Dynamix_LZW_decode( out_size, data_out );
	}

	//========================================
	//========================================
	//========================================

	sprintf( str, "%s", name );
	fp_out = fopen( str, "wb" );
	if( !fp_out ) return;


	fwrite( &data_out, 1, out_size, fp_out );
	fclose(fp_out);
}



void Dump_File_Rmf_Sub( FILE *fp_in, char *name, char *ext, UINT32 in_size, bool full_dump )
{
	FILE *fp_out;
	char type[4+1];
	UINT32 lcv, seek_start, out_size;
	bool raw_file;
	UINT32 multi_sng, multi_lcv;


	seek_start = ftell(fp_in);

	type[4] = 0;
	raw_file = false;
	multi_sng = 0;



	// find uncompressed size
	out_size = 0;
	lcv = 0;
	while( lcv < in_size ) {
		UINT32 pack_size;
		UINT8 flag;
		bool packed;


		/*
		1- ID (4 bytes)
		2- Size (3 bytes)
		3- Flag (1 byte)
		*/

		fread( &type, 1, 4, fp_in );

		lcv += 4;
		out_size += 4;



		// Adventures of Willy Beamish (Floppy)
		// Betrayal at Krondor (CD)
		if( type[3] != ':' ) {
			raw_file = true;

			out_size = in_size;
			break;
		}



		pack_size = 0;
		fread( &pack_size, 1, 3, fp_in );
		fread( &flag, 1, 1, fp_in );

		lcv += 4;
		out_size += 4;



		packed = false;

		if(0) {}

		else if( strcmp( ext, "ADS" ) == 0 ) {
			if(0) {}
			else if( strcmp( type, "SCR:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "BMP" ) == 0 ) {
			if(0) {}
			else if( strcmp( type, "BIN:" ) == 0 ) packed = true;
			else if( strcmp( type, "VGA:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "DDS" ) == 0 ) {
			if( strcmp( type, "DDS:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "OVL" ) == 0 ) {
			if(0) {}
			else if( strcmp( type, "ADL:" ) == 0 ) packed = true;
			else if( strcmp( type, "ADS:" ) == 0 ) packed = true;
			else if( strcmp( type, "APA:" ) == 0 ) packed = true;
			else if( strcmp( type, "ASB:" ) == 0 ) packed = true;
			else if( strcmp( type, "GMD:" ) == 0 ) packed = true;
			else if( strcmp( type, "M32:" ) == 0 ) packed = true;
			else if( strcmp( type, "NLD:" ) == 0 ) packed = true;
			else if( strcmp( type, "PRO:" ) == 0 ) packed = true;
			else if( strcmp( type, "PS1:" ) == 0 ) packed = true;
			else if( strcmp( type, "SBL:" ) == 0 ) packed = true;
			else if( strcmp( type, "SBP:" ) == 0 ) packed = true;
			else if( strcmp( type, "STD:" ) == 0 ) packed = true;
			else if( strcmp( type, "TAN:" ) == 0 ) packed = true;
			else if( strcmp( type, "001:" ) == 0 ) packed = true;
			else if( strcmp( type, "003:" ) == 0 ) packed = true;
			else if( strcmp( type, "004:" ) == 0 ) packed = true;
			else if( strcmp( type, "101:" ) == 0 ) packed = true;


			else if( strcmp( type, "VGA:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "SDS" ) == 0 ) {
			if( strcmp( type, "SDS:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "SNG" ) == 0 ) {
			if( strcmp( type, "SNG:" ) == 0 ) packed = true;


			// WARNING: Heart of China = multi-SNG resources
			if( packed ) multi_sng++;
		}

		else if( strcmp( ext, "TDS" ) == 0 ) {
			if( strcmp( type, "TDS:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "TTM" ) == 0 ) {
			if( strcmp( type, "TT3:" ) == 0 ) packed = true;
		}


		
		if( packed == true ) {
			UINT8 method;
			UINT32 unpack_size;


			fread( &method, 1, 1, fp_in );
			fread( &unpack_size, 1, 4, fp_in );
			pack_size -= (1 + 4);


			lcv += (1 + 4 + pack_size);
			out_size += (1 + 4 + unpack_size);
		}

		else if( flag == 0x80 ) {
			pack_size = 0;
		}

		else {
			lcv += pack_size;
			out_size += pack_size;
		}


		fseek( fp_in, pack_size, SEEK_CUR );
	}

	//========================================
	//========================================
	//========================================

	multi_lcv = 0;


	fp_out = fopen( name, "wb" );
	if( !fp_out ) return;


	fseek( fp_in, seek_start, SEEK_SET );



	// raw format
	if( raw_file == true ) {
		fread( &data_out, 1, in_size, fp_in );
		fwrite( &data_out, 1, in_size, fp_out );


		fclose( fp_out );
		return;
	}



	lcv = 0;
	while( out_size > 0 ) {
		UINT32 pack_size;
		UINT8 flag;
		bool packed;


		/*
		1- ID (4 bytes)
		2- Size (3 bytes)
		3- Flag (1 byte)
		*/

		fread( &type, 1, 4, fp_in );


		// Adventures of Willy Beamish (Floppy)
		// Betrayal at Krondor (CD)
		if( type[3] != ':' ) {
			out_size = in_size;
			break;
		}


		pack_size = 0;
		fread( &pack_size, 1, 3, fp_in );
		fread( &flag, 1, 1, fp_in );



		packed = false;
		if(0) {}

		else if( strcmp( ext, "ADS" ) == 0 ) {
			if(0) {}
			else if( strcmp( type, "SCR:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "BMP" ) == 0 ) {
			if(0) {}
			else if( strcmp( type, "BIN:" ) == 0 ) packed = true;
			else if( strcmp( type, "VGA:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "DDS" ) == 0 ) {
			if( strcmp( type, "DDS:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "OVL" ) == 0 ) {
			if(0) {}
			else if( strcmp( type, "ADL:" ) == 0 ) packed = true;
			else if( strcmp( type, "ADS:" ) == 0 ) packed = true;
			else if( strcmp( type, "APA:" ) == 0 ) packed = true;
			else if( strcmp( type, "ASB:" ) == 0 ) packed = true;
			else if( strcmp( type, "GMD:" ) == 0 ) packed = true;
			else if( strcmp( type, "M32:" ) == 0 ) packed = true;
			else if( strcmp( type, "NLD:" ) == 0 ) packed = true;
			else if( strcmp( type, "PRO:" ) == 0 ) packed = true;
			else if( strcmp( type, "PS1:" ) == 0 ) packed = true;
			else if( strcmp( type, "SBL:" ) == 0 ) packed = true;
			else if( strcmp( type, "SBP:" ) == 0 ) packed = true;
			else if( strcmp( type, "STD:" ) == 0 ) packed = true;
			else if( strcmp( type, "TAN:" ) == 0 ) packed = true;
			else if( strcmp( type, "001:" ) == 0 ) packed = true;
			else if( strcmp( type, "003:" ) == 0 ) packed = true;
			else if( strcmp( type, "004:" ) == 0 ) packed = true;
			else if( strcmp( type, "101:" ) == 0 ) packed = true;


			else if( strcmp( type, "VGA:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "SDS" ) == 0 ) {
			if( strcmp( type, "SDS:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "SNG" ) == 0 ) {
			if( strcmp( type, "SNG:" ) == 0 ) packed = true;


			if( packed == true ) multi_lcv++;
			if( multi_lcv > 1 && full_dump == false ) {
				char str[256], str2[256];
				int len;


				len = strlen(name);
				
				strcpy( str2, name );
				str2[ len-4 ] = 0;

				sprintf( str, "%s-%d.SND", str2, multi_lcv );
				fclose( fp_out );


				fp_out = fopen( str, "wb" );
				if( !fp_out ) return;
			}
		}

		else if( strcmp( ext, "TDS" ) == 0 ) {
			if( strcmp( type, "TDS:" ) == 0 ) packed = true;
		}

		else if( strcmp( ext, "TTM" ) == 0 ) {
			if( strcmp( type, "TT3:" ) == 0 ) packed = true;
		}



		if( packed == true ) {
			UINT8 method;
			UINT32 unpack_size;


			fread( &method, 1, 1, fp_in );
			fread( &unpack_size, 1, 4, fp_in );
			pack_size -= (1 + 4);


			reset_file( data_in, pack_size );
			fread( &data_in, 1, pack_size, fp_in );


			if( method == 1 )
				Dynamix_RLE_decode( unpack_size, data_out );

			else if( method == 2 )
				Dynamix_LZW_decode( unpack_size, data_out );

			// ------------------------------------------------
			// ------------------------------------------------

			method = 0;
			pack_size = unpack_size + (1 + 4);


			if( full_dump == true ) {
				fwrite( &type, 1, 4, fp_out );
				fwrite( &pack_size, 1, 3, fp_out );
				fwrite( &flag, 1, 1, fp_out );

				fwrite( &method, 1, 1, fp_out );
				fwrite( &unpack_size, 1, 4, fp_out );
			}

			fwrite( &data_out, 1, unpack_size, fp_out );


			out_size -= (4 + 3 + 1 + 1 + 4 + unpack_size );
		}

		else if( flag == 0x80 ) {
			out_size -= (4 + 3 + 1);
			pack_size = out_size;


			if( full_dump == true ) {
				fwrite( &type, 1, 4, fp_out );
				fwrite( &pack_size, 1, 3, fp_out );
				fwrite( &flag, 1, 1, fp_out );
			}
		}

		else {
			fread( &data_out, 1, pack_size, fp_in );


			if( full_dump == true ) {
				fwrite( &type, 1, 4, fp_out );
				fwrite( &pack_size, 1, 3, fp_out );
				fwrite( &flag, 1, 1, fp_out );
				fwrite( &data_out, 1, pack_size, fp_out );
			}


			out_size -= (4 + 3 + 1 + pack_size );
		}
	}



	fclose( fp_out );
}



void Dynamix_Scan_SND( int mode )
{
	UINT32 in_ptr;
	UINT32 map_cnt;
	UINT16 map_total;
	char str_id[4+1];


	in_ptr = 8;
	map_cnt = 0;
	map_total = 65535;
	str_id[4] = 0;


	while( map_cnt < map_total ) {
		if( mode == 1 )
			in_ptr = ftell(fp);

		else if( mode == 2 )
			fseek( fp, in_ptr, SEEK_SET );


		fread( &str_id, 1, 4, fp );


		if( strcmp( str_id, "TAG:" ) == 0 ) {
			// file strings
			UINT32 res_size;
			UINT16 num_files;


			/*
			4-byte header size
			*/

			fread( &res_size, 1, 4, fp );
			fread( &num_files, 1, 2, fp );


			for( int lcv=0; lcv<num_files; lcv++ ) {
				UINT16 index;
				UINT16 lcv2;


				fread( &index, 1, 2, fp );

				lcv2 = 0;
				while(1) {
					char ch;

					ch = fgetc(fp);
					sound_map[index].name[lcv2++] = ch;

					if( ch == 0 ) break;
				}
			}


			map_total = num_files;
		}

		else if( strcmp( str_id, "FNM:" ) == 0 ) {
			// file names
			UINT32 res_size;


			/*
			4-byte header size
			2-byte # strings
			<...>
			*/

			fread( &res_size, 1, 4, fp );


			fseek( fp, res_size, SEEK_CUR );
		}

		else if( strcmp( str_id, "INF:" ) == 0 ) {
			UINT32 map_size;
			UINT16 dummy1;
			UINT8 dummy2;


			// 4-byte packed size (after entry)
			fread( &map_size, 1, 4, fp );

				
			if( mode == 1 ) {
				/*
				header1:
				2-byte index #
				*/

				for( int lcv=0; lcv < map_size / 2; lcv++ ) {
					fread( &sound_map[lcv].index_num, 1, 2, fp );
				}
			}

			else if( mode == 2 ) {
				/*
				header1:
				2-byte ????
				2-byte # entries
				1-byte ????
				*/

				fread( &dummy1, 1, 2, fp );
				fread( &map_total, 1, 2, fp );
				fread( &dummy2, 1, 1, fp );


				/*
				table:
				2-byte index #
				4-byte absolute ptr (after this entry)
				*/

				for( int lcv=0; lcv < map_total; lcv++ ) {
					fread( &sound_map[lcv].index_num, 1, 2, fp );
					fread( &sound_map[lcv].data_ptr, 1, 4, fp );
				}
			}
		}

		else if( strcmp( str_id, "DAT:" ) == 0 ) {
			UINT32 in_size, out_size;
			UINT16 index, dummy1;
			UINT8 compression;

			/*
			4-byte packed size (after this entry)

			2-byte index #
			2-byte ??? resource type (0009 = midi)
			1-byte ??? compression (00 = raw)
			4-byte output size (after this entry)
			*/

			fread( &in_size, 1, 4, fp );
			fread( &index, 1, 2, fp );
			fread( &dummy1, 1, 2, fp );
			fread( &compression, 1, 1, fp );
			fread( &out_size, 1, 4, fp );


			in_size -= (2 + 2 + 1 + 4);


			if( mode == 1 ) {
				Dump_File_Ext( compression, fp, in_size, out_size, sound_map[index].name, "SND" );


				printf( "%d = %s.SND | Offset = %X\n",
					index,
					sound_map[index].name,
					in_ptr );
			}

			else if( mode == 2 ) {
				Dump_File( compression, fp, in_size, out_size, index, "SND" );


				printf( "%d.SND | Offset = %X\n",
					index,
					in_ptr );
			}


			map_cnt++;
		}


		in_ptr = sound_map[map_cnt].data_ptr;
	}
}



void Dynamix_Scan_SSM( UINT32 stop )
{
	UINT32 in_ptr;
	char str_id[4+1];


	in_ptr = 8;
	str_id[4] = 0;


	while( in_ptr < stop ) {
		UINT32 in_size, out_size;
		UINT8 compression;


		/*
		4-byte ID
		4-byte packed size (after this entry)
		1-byte compression (00 = raw)
		4-byte output size (after this entry)
		*/

		fseek( fp, in_ptr, SEEK_SET );
		fread( &str_id, 1, 4, fp );
		fread( &in_size, 1, 4, fp );
		fread( &compression, 1, 1, fp );
		fread( &out_size, 1, 4, fp );


		str_id[3] = 0;


		Dump_File_Ext( compression, fp, in_size-5, out_size, str_id, "DRV" );


		printf( "%s.DRV | Offset = %X\n",
			str_id,
			in_ptr );


		in_ptr += (4 + 4 + in_size);
	}
}



void Dynamix_Scan_RMF( char *type )
{
	UINT32 version, mode;
	UINT16 num_volumes;


	/*
	00 01 06 07  (Adventures of Willy Beamish - Floppy)
	00 01 06 07  (Rise of the Dragon)
	00 02 04 07  (Heart of China)
	01 00 00 00  (Betrayal at Krondor - CD)
	*/

	fread( &version, 1, 4, fp );
	fread( &num_volumes, 1, 2, fp );

	if(0) {}
	else if( version == 0x07060100 ) mode = 1;
	else if( version == 0x07040200 ) mode = 2;
	else if( version == 0x00000001 ) mode = 3;



	res_cnt = 0;

	for( int lcv=0; lcv < num_volumes; lcv++ ) {
		UINT16 num_files;


		// Krondor CD
		if( feof(fp) ) break;


		/*
		1- name (13 bytes) (note: maybe 12 bytes + 1 ??? byte)
		2- files (2 bytes)
		*/

		fread( &rmf_names[lcv], 1, 13, fp );
		fread( &num_files, 1, 2, fp );



		for( int lcv2=0; lcv2 < num_files; lcv2++ ) {
			rmf_map[res_cnt].file_num = lcv;

			/*
			1- hash (4 bytes)
			2- offset (4 bytes)
			*/

			fread( &rmf_map[res_cnt].hash, 1, 4, fp );
			fread( &rmf_map[res_cnt].offset, 1, 4, fp );

			res_cnt++;
		}
	}

	fclose(fp);

	rmf_map[res_cnt].file_num = 0xff;

	//===================================================
	//===================================================
	//===================================================

	res_cnt = 0;
	lcv = 0xff;

	while( rmf_map[res_cnt].file_num != 0xff ) {
		UINT8 name[13+1];
		UINT32 in_size;
		int len;


		if( lcv != rmf_map[res_cnt].file_num ) {
			if( fp ) fclose(fp);
			lcv = rmf_map[res_cnt].file_num;


			if( rmf_names[lcv][0] == 0 ) break;


			fp = fopen( (char *) rmf_names[lcv], "rb" );
			if( !fp ) {
				lcv = 0xff;
				res_cnt++;

				continue;
			}
		}



		fseek( fp, rmf_map[res_cnt].offset, SEEK_SET );


		/*
		- name (13 bytes)
		- in_size (4 bytes)
		*/
		name[13] = 0;

		fread( &name, 1, 13, fp );
		fread( &in_size, 1, 4, fp );



		printf( "%s | File = %s   Offset = %X\n",
			name,
			rmf_names[lcv],
			rmf_map[res_cnt].offset );

		// ---------------------------------------
		// ---------------------------------------
		// ---------------------------------------

		len = strlen( (char *) name );


		// Normal file dumps
		if( stricmp( type, "all" ) == 0 ||
				stricmp( (char *) name + (len-3), type ) == 0 )
			Dump_File_Rmf_Sub( fp, (char *) name, (char *) name + (len-3), in_size, true );


		// MIDI file dumps only
		else if(
			stricmp( type, "snd" ) == 0 &&
			strcmp( (char *) name + (len-3), "SNG" ) == 0 )
		{
			char snd_name[80];

			strcpy( (char *) snd_name, (char *) name );
			snd_name[ len-1 ] = 'D';

			Dump_File_Rmf_Sub( fp, (char *) snd_name, (char *) name + (len-3), in_size, false );
		}


		res_cnt++;
	}


	fclose( fp );

	printf( "total = %d\n", res_cnt );
}



void Dump_Dynamix( int argc, char **argv )
{
	char str[4+1];
	UINT8 dummy;
	UINT16 ext;


	if( argc < 2+1 ) return;


	fp = fopen( argv[2], "rb" );
	if( !fp ) return;


	ext = strlen( argv[2] ) - 1;
	while( ext > 0 && argv[2][ext] != '.' ) {
		ext--;
	}


	// RMF
	if( stricmp( argv[2] + (ext+1), "RMF" ) == 0 )
		Dynamix_Scan_RMF( argv[3] );


	// Rise of the Dragon
	else if( stricmp( argv[2], "VOLUME.VGA" ) == 0 )
		Dynamix_Scan_RMF( argv[3] );



	/*
	Adventures of Willy Beamish
	- WILLYMUS.SX
	- WILLYSND.SX
	- Type 1

	
	Betrayal at Krondor
	- FRP.SX
	- SX.OVL
	- Type 2
	*/
	else if(
		stricmp( argv[2] + (ext+1), "SX" ) == 0 ||
		stricmp( argv[2] + (ext+1), "OVL" ) == 0 )
	{
		str[4] = 0;


		/*
		4-byte ID
		3-byte size (after full header)
		1-byte flag? ($80)
		*/

		fread( &str, 1, 4, fp );
		fread( &file_size, 1, 3, fp );
		fread( &dummy, 1, 1, fp );


		// add header
		file_size += 8;


		if( strcmp( str, "SND:" ) == 0 ) {
			UINT8 type;


			sscanf( argv[3], "%d", &type );
			Dynamix_Scan_SND(type);
		}

		else if( strcmp( str, "SSM:" ) == 0 )
			Dynamix_Scan_SSM( file_size );
	}
}
