#include <memory.h>
#include "types.h"



static UINT8 code_cur[256];
static UINT32 code_size;
static UINT32 code_len;


static struct {
	UINT8 str[256];
	UINT8 len;
} dict_table[0x4000];

static UINT32 dict_size;
static UINT32 dict_max;
static bool dict_full;



FILE *fp_out, *fp_in;





UINT32 Find_LZW_Code( UINT32 start )
{
	UINT32 find_last, find_len;


	// 1-byte code
	find_last = data_in[ start ];
	find_len = 2;


	while( start + ( find_len - 1 ) < in_size ) {
		bool hit;
		UINT32 lcv;


		hit = false;
		for( lcv=0x102; lcv<dict_size; lcv++ ) {
			// check lengths
			if( dict_table[lcv].len != find_len ) continue;


			// compare strings
			for( int lcv2=0; lcv2<dict_table[lcv].len; lcv2++ ) {
				if( dict_table[lcv].str[lcv2] != data_in[ start+lcv2 ] ) break;
			}


			// match! no extra entry needed
			if( lcv2 == dict_table[lcv].len && lcv2 > 0 ) {
				hit = true;
				break;
			}
		}



		// expand search
		if( hit == true ) {
			find_last = lcv;
			find_len++;

			continue;
		}

		break;
	}


	return find_last;
}


void Sierra_LZW_AGI_reset()
{
	memset( &dict_table, 0, sizeof(dict_table) );

	for( int lcv=0; lcv<256; lcv++ ) {
		dict_table[lcv].len = 1;
		dict_table[lcv].str[0] = lcv;
	}


	// 00-FF = ASCII
	// 100 = reset
	// 101 = stop
	dict_size = 0x102;
	dict_max = 0x200;
	dict_full = false;


	// start = 9 bit codes
	code_size = 9;
	code_len = 0;
}


void Sierra_LZW_AGI_packer( int mode, char *in, char *out )
{
	int lcv;

	
	fp_in = fopen( in, "rb" );
	if( !fp_in ) return;

	in_size = fread( &data_in, 1, 0x800000, fp_in );
	fclose( fp_in );

	// ------------------------------------------
	// ------------------------------------------

	reset_file_bits();


	lcv = 0;
	while( lcv < in_size ) {
		UINT32 new_code;


		// send reset
		if( lcv == 0 ) {
			Sierra_LZW_AGI_reset();

			pack_bits_right( code_size, 0x100 );
		}


		new_code = Find_LZW_Code( lcv );
		pack_bits_right( code_size, new_code );


		// expand string
		code_cur[ code_len++ ] = dict_table[ new_code ].str[0];
		lcv += dict_table[ new_code ].len;

		// -----------------------------------------------
		// -----------------------------------------------

		// add to dictionary: 2+ bytes only
		if( code_len >= 2 )
		{
			int lcv1;


			if( dict_full == false ) {
				// check full condition
				if( dict_size == dict_max && code_size == 12 ) {
					dict_full = true;

					lcv1 = dict_size;
				}
				else
					lcv1 = dict_size++;


				// expand dictionary (adaptive LZW)
				if( dict_size == dict_max && code_size < 12 ) {
					dict_max *= 2;
					code_size++;
				}


				// add new entry
				for( UINT32 lcv2=0; lcv2<code_len; lcv2++ ) {
					dict_table[lcv1].str[lcv2] = code_cur[lcv2];
					dict_table[lcv1].len++;
				}
			}


			
			// reset running code!
			for( lcv1=0; lcv1<dict_table[ new_code ].len; lcv1++ )
				code_cur[lcv1] = dict_table[ new_code ].str[lcv1];

			code_len = dict_table[ new_code ].len;



			// send reset code
			if( dict_size == dict_max && code_size == 12 ) {
				pack_bits_right( code_size, 0x100 );

				Sierra_LZW_AGI_reset();
			}
		}
	}


	pack_bits_right( code_size, 0x101 );
	flush_bits_right();

	// ------------------------------------------
	// ------------------------------------------

	fp_out = fopen( out, "wb" );
	if( !fp_out ) return;



	if( mode == 1 ) {
		// Sorcerian = 4-byte header (length)
		fwrite( &in_size, 1, 4, fp_out );
	}



	fwrite( &data_out, 1, out_size, fp_out );
	fclose( fp_out );
}
