/*
SCI0 - type 1 (MT32)




Channel = 1-16
Instruments = 1-128
Notes = 0-127
Velocity = 0-127


Special Adlib MIDI commands
- $4b = poly reserve
- $4e = set velocity


Ch16 = loop flags


Special analysis - MT32 vs Adlib
- unique parts
- unique instruments
*/



void Sierra_SCI0_1()
{
	int len;

	len = strlen(ffd.cFileName);


	if( len == 5+1+3 &&
			ffd.cFileName[0] == 's' &&
			ffd.cFileName[1] == 'o' &&
			ffd.cFileName[2] == 'u' &&
			ffd.cFileName[3] == 'n' &&
			ffd.cFileName[4] == 'd' ) {}

	else if(
			len == 5+1+3 &&
			ffd.cFileName[0] == 'S' &&
			ffd.cFileName[1] == 'O' &&
			ffd.cFileName[2] == 'U' &&
			ffd.cFileName[3] == 'N' &&
			ffd.cFileName[4] == 'D' ) {}

	else
		return;

	//*******************************************
	//*******************************************
	//*******************************************

	int sci_header;

	fprintf( fp_global, "%s\n", ffd.cFileName );



	fp = fopen( fileString, "rb" );
	if( !fp ) {
		return;
	}

	fread( &buffer, 1, BUF_TOTAL, fp );
	buf_size = ftell( fp );
	fclose(fp);


	// skip file header = '.snd'
	sci_header = 0;
	if( buffer[0] == 0x84 && buffer[1] == 0x00 ) sci_header = 2;


	len = strlen(fileString);


	
	// sound.003
	strcat( fileString, ".txt\0" );


	fp_log = fopen( fileString, "w" );
	if( !fp_log ) {
		return;
	}

	//**********************************************
	//**********************************************
	//**********************************************

	int warnings;
	int lcv, lcv2;


	buf_ptr = sci_header;
	warnings = 0;


	memset( chan_mt32, 0, sizeof(chan_mt32) );
	memset( chan_adlib, 0, sizeof(chan_adlib) );

	memset( patch_mt32, 0, sizeof(patch_mt32) );
	memset( patch_adlib, 0, sizeof(patch_adlib) );

	memset( perc_mt32, 0, sizeof(perc_mt32) );
	memset( perc_adlib, 0, sizeof(perc_adlib) );


	track_mt32 = 0;
	track_adlib = 0;

	//*****************************************
	//*****************************************
	//*****************************************

	UINT8 digital_pcm;


	memset( track_ofs, -1, sizeof(track_ofs) );
	memset( track_size, 0, sizeof(track_size) );



	// 00 = no, 02 = yes
	digital_pcm = UMEM8(BUF);
	buf_ptr++;

	if( digital_pcm ) {
		fprintf( fp_log, "Digital PCM = yes\n\n\n" );


		// unhandled case!!
		fprintf( fp_global, "Digital PCM = yes\n\n\n" );
		warnings++;
	}



	// cycle through track info
	for( lcv=0; lcv<16; lcv++ ) {
		UINT8 voice, device;

		voice = ( UMEM8(BUF) & 0xf0 ) >> 4;
		device = UMEM8(BUF) & 0x0f;
		buf_ptr++;



		fprintf( fp_log, "ch%d = ", voice );
		if( device & 1 ) { fprintf( fp_log, "Adlib  " ); track_adlib = 1; }
		if( device & 2 ) fprintf( fp_log, "PCjr  " );
		if( device & 4 ) fprintf( fp_log, "Control  " );
		if( lcv >= 1 && lcv <= 9 ) fprintf( fp_log, "MT-32" ); track_mt32 = 1;


		fprintf( fp_log, "\n" );
	}


	fprintf( fp_log, "\n" );
	fprintf( fp_log, "\n" );
	fprintf( fp_log, "\n" );

	//*****************************************
	//*****************************************
	//*****************************************

	// loop through data
	int cmd, arg1, arg2, chan;
	int pcm_panpot;
	bool run_cmd;
	bool play_midi;



	pcm_panpot = 0x40;


	if( track_mt32 ) {
		bool hit = false;


		if( track_adlib == 0 && track_mt32 == 1 ) {
			fprintf( fp_global, "[%06X]  Play part %d = MT-32", buf_ptr, lcv+1 );

			hit = true;
		}


		if( hit ) fprintf( fp_global, "\n" );


		global_notice++;
	}
	else if( track_adlib ) {
		fprintf( fp_global, "[%06X]  Play part %d = Adlib\n", buf_ptr, lcv+1 );


		global_notice++;
	}



	if( digital_pcm == 2 ) {
		int freq;


		// header bytes
		fprintf( fp_log, "Header bytes\n" );

	
		fprintf( fp_log, "[%06X]  ", buf_ptr );
		for( int lcv_buf=0; lcv_buf<2; lcv_buf++ ) {
			fprintf( fp_log, "%02X ", UMEM8(BUF) );

			buf_ptr++;
		}
		fprintf( fp_log, "\n" );


		freq = UMEM16(BUF);
		fprintf( fp_log, "- Frequency = %d\n", freq );



		fprintf( fp_log, "- *** Digital PCM!!\n" );
		fprintf( fp_global, "[%06X]  *** Digital PCM = %d!!\n", buf_ptr-2, freq );


		if( track_mt32 == 0 ) {
			fprintf( fp_global, "- MT-32 not found!!! (@@@)\n" );
			global_notice++;
		}

		if( pcm_panpot != 0x40 ) {
			fprintf( fp_global, "- MIDI panpot = %X!! (@@@)\n", pcm_panpot );
			global_notice++;
		}


		global_notice++;
	}

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	run_cmd = false;
	play_midi = true;

	while( play_midi ) {
		int start_ptr;


		// ABORT!! - some type of error
		if( buf_ptr >= buf_size ) {
			fprintf( fp_log, "ERROR - no EOF detected!\n" );
			fprintf( fp_global, "ERROR - no EOF detected!\n" );


			warnings++;
			global_notice++;

			break;
		}



		// special delay
		if( UMEM8(BUF) == 0xf8 ) {
			// header
			fprintf( fp_log, "[%06X]  ", buf_ptr );

				
			while( UMEM8(BUF) == 0xf8 ) {
				fprintf( fp_log, "{%02X} ", UMEM8(BUF) );

				buf_ptr++;
			}

				
			fprintf( fp_log, "\n" );
		}

		//------------------------------------------
		//------------------------------------------
		//------------------------------------------

		start_ptr = buf_ptr;


		// header
		fprintf( fp_log, "[%06X]  ", buf_ptr );


		// delay byte
		fprintf( fp_log, "{%02X} ", UMEM8(BUF) );

		buf_ptr++;



		// new MIDI command
		if( UMEM8(BUF) >= 0x80 ) {
			cmd = UMEM8(BUF);
			buf_ptr++;

			chan = cmd & 0xf;
			run_cmd = false;
		}
		else {
			fprintf( fp_log, " --  " );

			run_cmd = true;
		}


		arg1 = UMEM8(BUF);
		arg2 = UMEM8(BUF+1);


		switch( cmd >> 4 ) {
			case 0x8:
			case 0x9:
			case 0xa:
			case 0xb:
			case 0xe:
				if( (cmd >> 4) == 0xb && arg1 == 0xa && pcm_panpot == 0x40 ) {
					if( track_mt32 )
						pcm_panpot = arg2;
				}


				buf_ptr += 2;


				if( run_cmd == false )
					fprintf( fp_log, "[%02X] %02X %02X\n", cmd, arg1, arg2 );
				else
					fprintf( fp_log, "%02X %02X\n", arg1, arg2 );
				break;


			case 0xc:
			case 0xd:
				buf_ptr += 1;


				if( run_cmd == false )
					fprintf( fp_log, "[%02X] %02X\n", cmd, arg1 );
				else
					fprintf( fp_log, "%02X\n", arg1 );
				break;


			case 0xf:
				switch( cmd ) {
					case 0xfc:
						fprintf( fp_log, "[%02X]  (STOP)\n", cmd );

						cmd = 0x00;
						play_midi = false;
						break;


					default:
						if( run_cmd == false ) {
							fprintf( fp_log, "[%02X]  (Unknown)\n", cmd );
							fprintf( fp_global, "[%02X]  (Unknown)\n", cmd );
						}
						else {
							fprintf( fp_log, "%02X  (Unknown)\n", arg1 );
							fprintf( fp_global, "%02X  (Unknown)\n", arg1 );
						}


						warnings++;
						global_notice++;
						break;
				}
				break;


			default:
				buf_ptr++;
				warnings++;
				global_notice++;


				if( run_cmd == false ) {
					fprintf( fp_log, "%02X = (Unknown)  (WARN!)\n", cmd );
					fprintf( fp_global, "%02X = (Unknown)  (WARN!)\n", cmd );
				}
				else {
					fprintf( fp_log, "%02X  (WARN!)\n", arg1 );
					fprintf( fp_global, "%02X  (WARN!)\n", arg1 );
				}
				break;
		}

		//------------------------------------------
		//------------------------------------------
		//------------------------------------------

		// look for soundtrack quirks
		// - NOTE: Sierra ch16 = special loop flags

		if( (track_mt32 || track_adlib) && chan != 16-1 ) {
			if( (cmd >> 4) == 0x9 && arg2 > 0 ) {
				if( track_mt32 ) chan_mt32[ chan ] = 1;
				if( track_adlib ) chan_adlib[ chan ] = 1;
			}


			if( (cmd >> 4) == 0xc ) {
				if( chan != 10-1 ) {
					if( track_mt32 ) patch_mt32[ chan ][ arg1 ] = 1;
					if( track_adlib ) patch_adlib[ chan ][ arg1 ] = 1;
				}
				else {
					if( track_mt32 ) perc_mt32[ chan ][ arg1 ] = 1;
					if( track_adlib ) perc_adlib[ chan ][ arg1 ] = 1;
				}
			}


			if( (cmd >> 4) == 0xd && arg1 > 0 ) {
				fprintf( fp_global, "[%06X]  Channel aftertouch = %02X %02X  (@@@)\n", start_ptr, cmd, arg1 );
				global_notice++;
			}


			if( (cmd >> 4) == 0xa && arg2 > 0 ) {
				fprintf( fp_global, "[%06X]  Polyphonic aftertouch = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );
				global_notice++;
			}


			if( cmd == 0xc9 && arg1 > 0 ) {
				fprintf( fp_global, "[%06X]  *** Drum kit = %02X %02X  (@@@)\n", start_ptr, cmd, arg1 );
				global_notice++;
			}


			if( cmd == 0x99 && arg2 > 0 ) {
				bool hit = false;
				bool hit2a = false;
				bool hit2b = false;
				bool hit3 = false;



				if( track_mt32 ) {
					// CM32L + MT-32 custom
					if( arg1 >= 76 && arg1 <= 87 ) hit2a = true;

					// CM32L only
					if( arg1 >= 88 && arg1 <= 108 ) hit2b = true;
				}


				// Unknown area
				if( track_mt32 ) {
					if( arg1 < 27 || arg1 > 108 ) hit3 = true;
				}


				if( hit ) fprintf( fp_global, "[%06X]  (GS) Percussion = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );
				if( hit2a ) fprintf( fp_global, "[%06X]  (MT-32 timbre / LAPC-I sound) Percussion = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );
				if( hit2b ) fprintf( fp_global, "[%06X]  (LAPC-I) Percussion = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );
				if( hit3 ) fprintf( fp_global, "[%06X]  (UNK) Percussion = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );


				if( hit ) global_notice++;
				if( hit2a ) global_notice++;
				if( hit2b ) global_notice++;
				if( hit3 ) global_notice++;
			}


			if( (cmd>>4) == 0xb ) {
				bool hit, hit2;


				hit = true;
				hit2 = false;


				if( arg1 == 0x01 ) hit = false;
				if( arg1 == 0x07 ) hit = false;
				if( arg1 == 0x0a ) hit = false;
				if( arg1 == 0x40 ) hit = false;
				if( arg1 == 0x7b ) hit = false;


				// log modulation
				if( arg1 == 0x01 && arg2 != 0x00 ) hit2 = true;


				if( hit && fp_global ) {
					bool hit_adlib = false;


					if( arg1 == 0x4b ) hit_adlib = true;
					if( arg1 == 0x4e ) hit_adlib = true;


					if( hit_adlib == false ) {
						fprintf( fp_global, "[%06X]  Control change = %02X %02X %02X  (*** WARN!!! ***)\n", start_ptr, cmd, arg1, arg2 );

						warnings++;
					}
					else
						fprintf( fp_global, "[%06X]  Control change = %02X %02X %02X  (Adlib)\n", start_ptr, cmd, arg1, arg2 );


					global_notice++;
				}

				else if( hit2 && fp_global )
					fprintf( fp_global, "[%06X]  Control change = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );
			}
		}
	} // end play midi


	fprintf( fp_log, "\n" );
	fprintf( fp_log, "\n" );
	fprintf( fp_log, "\n" );


	if( warnings ) fprintf( fp_log, "Warnings = %d", warnings );

	fclose(fp_log);

	//************************************************
	//************************************************
	//************************************************

	// look for soundtrack quirks


	bool header;


	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		if( chan_mt32[lcv] == 1 && (lcv==1-1 || lcv>=11-1) ) {
			if( header == false ) {
				fprintf( fp_global, "*** MT-32 = " );

				header = true;
			}


			fprintf( fp_global, "ch%d (@@@) ", lcv+1 );
			global_notice++;
		}

		else if( chan_mt32[lcv] == 1 && chan_adlib[lcv] == 0 ) {
			if( header == false ) {
				fprintf( fp_global, "*** MT-32 = " );

				header = true;
			}


			fprintf( fp_global, "ch%d ", lcv+1 );
			global_notice++;
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		for( lcv2=0; lcv2<128; lcv2++ ) {
			if( patch_adlib[lcv][lcv2] == 0 && patch_mt32[lcv][lcv2] == 1 ) {
				if( header == false ) {
					fprintf( fp_global, "*** MT-32 patch (1-128)\n" );

					header = true;
				}


				fprintf( fp_global, "- ch%d = [%02X] %d\n", lcv+1, lcv2, lcv2+1 );


				global_notice++;
			}
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );



	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		for( lcv2=0; lcv2<128; lcv2++ ) {
			if( perc_adlib[lcv][lcv2] == 0 && perc_mt32[lcv][lcv2] == 1 ) {
				if( header == false ) {
					fprintf( fp_global, "*** MT-32 percussion (0-127)\n" );

					header = true;
				}


				fprintf( fp_global, "- ch%d = [%02X] %d\n", lcv+1, lcv2, lcv2 );


				global_notice++;
			}
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		if( chan_adlib[lcv] == 1 && chan_mt32[lcv] == 0 ) {
			if( header == false ) {
				fprintf( fp_global, "*** Adlib = " );

				header = true;
			}


			fprintf( fp_global, "ch%d ", lcv+1 );
			global_notice++;
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		for( lcv2=0; lcv2<128; lcv2++ ) {
			if( patch_adlib[lcv][lcv2] == 1 && patch_mt32[lcv][lcv2] == 0 ) {
				if( header == false ) {
					fprintf( fp_global, "*** Adlib patch (1-128)\n" );

					header = true;
				}


				fprintf( fp_global, "- ch%d = [%02X] %d\n", lcv+1, lcv2, lcv2+1 );


				global_notice++;
			}
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );



	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		for( lcv2=0; lcv2<128; lcv2++ ) {
			if( perc_adlib[lcv][lcv2] == 1 && perc_mt32[lcv][lcv2] == 0 ) {
				if( header == false ) {
					fprintf( fp_global, "*** Adlib percussion (0-127)\n" );

					header = true;
				}


				fprintf( fp_global, "- ch%d = [%02X] %d\n", lcv+1, lcv2, lcv2 );


				global_notice++;
			}
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	global_warnings += warnings;


	fprintf( fp_global, "\n" );
}
