/*
SCI1 - type 3 (MT32 + GM)




Channel = 1-16
Instruments = 1-128
Notes = 0-127
Velocity = 0-127


Special Adlib MIDI commands
- $4b = poly reserve
- $4e = set velocity


Ch16 = loop flags


Special analysis - GM vs MT32
- unique parts
- unique instruments
*/



void Sierra_SCI1_3()
{
	int len;

	len = strlen(ffd.cFileName);


	if( len > 4 &&
			ffd.cFileName[len-4] == '.' &&
			ffd.cFileName[len-3] == 's' &&
			ffd.cFileName[len-2] == 'n' &&
			ffd.cFileName[len-1] == 'd' ) {}
	
	else if(
			len > 4 &&
			ffd.cFileName[len-4] == '.' &&
			ffd.cFileName[len-3] == 'S' &&
			ffd.cFileName[len-2] == 'N' &&
			ffd.cFileName[len-1] == 'D' ) {}
	
	else
		return;

	//*******************************************
	//*******************************************
	//*******************************************

	int sci_header;


	fprintf( fp_global, "%s\n", ffd.cFileName );



	fp = fopen( fileString, "rb" );
	if( !fp ) {
		return;
	}

	fread( &buffer, 1, BUF_TOTAL, fp );
	buf_size = ftell( fp );
	fclose(fp);


	// skip file header = '.snd'
	sci_header = 0;
	if( buffer[0] == 0x84 && buffer[1] == 0x00 ) sci_header = 2;



	len = strlen(fileString);


	// 0.snd
	fileString[len-3] = 't';
	fileString[len-2] = 'x';
	fileString[len-1] = 't';


	fp_log = fopen( fileString, "w" );
	if( !fp_log ) {
		return;
	}

	//**********************************************
	//**********************************************
	//**********************************************

	// 1 = MT-32
	fp = fopen( "1.pat", "rb" );
	if( fp ) {
		fseek( fp, 0+2, SEEK_SET );
		fread( &map_patch_mt32, 1, 128, fp );

		fseek( fp, 0x180+2, SEEK_SET );
		fread( &map_perc_mt32, 1, 128, fp );

		fclose(fp);
	}



	// 4 = GM
	fp = fopen( "4.pat", "rb" );
	if( fp ) {
		fseek( fp, 0+2, SEEK_SET );
		fread( &map_patch_gm, 1, 128, fp );

		fseek( fp, 0x180+2, SEEK_SET );
		fread( &map_perc_gm, 1, 128, fp );

		fclose(fp);
	}

	//**********************************************
	//**********************************************
	//**********************************************

	int warnings;
	int lcv, lcv2;


	buf_ptr = sci_header;
	warnings = 0;


	memset( chan_gm, 0, sizeof(chan_gm) );
	memset( chan_mt32, 0, sizeof(chan_mt32) );
	memset( chan_adlib, 0, sizeof(chan_adlib) );

	memset( patch_gm, 0, sizeof(patch_gm) );
	memset( patch_mt32, 0, sizeof(patch_mt32) );
	memset( patch_adlib, 0, sizeof(patch_adlib) );

	memset( perc_gm, 0, sizeof(perc_gm) );
	memset( perc_mt32, 0, sizeof(perc_mt32) );
	memset( perc_adlib, 0, sizeof(perc_adlib) );



	// sysex
	if( UMEM8(BUF) == 0xf0 ) {
		fprintf( fp_log, "[000000]  SysEx transfer = %d bytes\n", UMEM8(BUF+1) );
		fprintf( fp_global, "[000000]  SysEx transfer = %d bytes\n", UMEM8(BUF+1) );
		global_notice++;


		buf_ptr += 2;


		// unknown - ignored by interpreter (F0 case)
		buf_ptr += 6;

	
		fprintf( fp_log, "\n" );
		fprintf( fp_log, "\n" );
	}

	//*****************************************
	//*****************************************
	//*****************************************

	memset( track_ofs, -1, sizeof(track_ofs) );
	memset( track_size, 0, sizeof(track_size) );


	
	// cycle through track info
	while( UMEM8(BUF) != 0xff ) {
		int track_id;
		int part;


		// track ID
		track_id = UMEM8(BUF);
		buf_ptr++;


		// log
		fprintf( fp_log, "Track %d = ", track_id );

		switch( track_id ) {
		case 0: fprintf( fp_log, "Adlib, Soundblaster\n" ); break;
		case 7: fprintf( fp_log, "General MIDI\n" ); break;
		case 9: fprintf( fp_log, "CMS\n" ); break;
		case 12: fprintf( fp_log, "MT-32\n" ); break;
		case 18: fprintf( fp_log, "PC Speaker\n" ); break;
		case 19: fprintf( fp_log, "Tandy 1000, PS/1\n" ); break;
		default: fprintf( fp_log, "Unknown\n" ); break;
		}



		// track offsets
		part = 0;

		while( UMEM8(BUF) != 0xff ) {
			// stop code
			buf_ptr++;


			// arg1 = dummy byte
			if( UMEM8(BUF) != 0 ) {
				fprintf( fp_log, "[%06X]  WARN: Unknown track arg1 = %d\n", buf_ptr, UMEM8(BUF) );
				fprintf( fp_global, "[%06X]  WARN: Unknown track arg1 = %d\n", buf_ptr, UMEM8(BUF) );


				warnings++;
				global_notice++;
			}
			buf_ptr++;


			// arg2 = 16-bit offset
			track_ofs[track_id][part] = UMEM16(BUF) + sci_header;
			buf_ptr += 2;


			// arg3 = 16-bit size
			track_size[track_id][part] = UMEM16(BUF);
			buf_ptr += 2;


			part++;
		}


		// skip stop code
		buf_ptr++;

	
	
		fprintf( fp_log, "- Play parts = %d\n", part );
		fprintf( fp_log, "\n" );
	}


	// skip stop code
	buf_ptr++;


	fprintf( fp_log, "\n" );
	fprintf( fp_log, "\n" );
	fprintf( fp_log, "\n" );
	fprintf( fp_log, "\n" );

	//*****************************************
	//*****************************************
	//*****************************************

	memset( part_list, 0, sizeof(part_list) );
	part_size = 0;



	// now sort track offsets
	for( int parts=0; parts<256; parts++ ) {
		bool found = false;
		int low_offset = 0x7fffffff;

		
		for( int lcv1=0; lcv1<256; lcv1++ ) {
			for( int lcv2=0; lcv2<256; lcv2++ ) {
				if( track_size[lcv1][lcv2] == 0 ) break;

				if( track_ofs[lcv1][lcv2] == 0 ) continue;
				if( track_ofs[lcv1][lcv2] >= low_offset ) continue;

				if( part_size > 0 && track_ofs[lcv1][lcv2] <= part_list[part_size-1] ) continue;


				low_offset = track_ofs[lcv1][lcv2];
				found = true;
			}
		}

	
		if( found == true ) {
			part_list[part_size] = low_offset;


			part_size++;
		}
		else
			break;
	}

	//*****************************************
	//*****************************************
	//*****************************************

	// loop through data
	int cmd, arg1, arg2, chan;
	int pcm_panpot;
	bool run_cmd;
	bool play_midi;



	pcm_panpot = 0x40;


	// run through tracks
	for( lcv=0; lcv<part_size; lcv++ ) {
		bool digital_pcm;


		buf_ptr = part_list[lcv];


		track_gm = 0;
		track_mt32 = 0;
		track_adlib = 0;

		digital_pcm = false;


		// debug log
		fprintf( fp_log, "Play part %d\n", lcv+1 );

		for( int lcv2=0; lcv2<256; lcv2++ ) {
			for( int lcv3=0; lcv3<256; lcv3++ ) {
				if( track_size[lcv2][lcv3] == 0 ) break;

				if( (UINT32) track_ofs[lcv2][lcv3] == buf_ptr ) {
					if( UMEM8(BUF) == 0xfe && UMEM8(BUF+1) == 0x00 ) digital_pcm = true;

					
					switch( lcv2 ) {
					case 0:
						if( digital_pcm == true ) fprintf( fp_log, "- Soundblaster\n" );
						else fprintf( fp_log, "- Adlib\n" );
						track_adlib = 1; break;
					case 7: fprintf( fp_log, "- General MIDI\n" ); track_gm = 1; break;
					case 9: fprintf( fp_log, "- CMS\n" ); break;
					case 12: fprintf( fp_log, "- MT-32\n" ); track_mt32 = 1; break;
					case 18: fprintf( fp_log, "- PC Speaker\n" ); break;
					case 19: fprintf( fp_log, "- Tandy 1000\n" ); break;
					default: fprintf( fp_log, "- Unknown %d\n", lcv2 ); break;
					}

					break;
				}
			}
		}

		fprintf( fp_log, "\n" );
		fprintf( fp_log, "\n" );



		if( track_gm || track_mt32 ) {
			bool hit = false;


			if( track_gm == 1 && track_mt32 == 0 ) {
				fprintf( fp_global, "[%06X]  Play part %d = GM", buf_ptr, lcv+1 );

				hit = true;
			}

			if( track_gm == 0 && track_mt32 == 1 ) {
				fprintf( fp_global, "[%06X]  Play part %d = MT-32", buf_ptr, lcv+1 );

				hit = true;
			}


			if( hit && track_adlib == 1 ) fprintf( fp_global, " Adlib" );
			if( hit ) fprintf( fp_global, "\n" );


			global_notice++;
		}
		else if( track_adlib ) {
			fprintf( fp_global, "[%06X]  Play part %d = Adlib\n", buf_ptr, lcv+1 );


			global_notice++;
		}



		if( digital_pcm == true ) {
			int freq;


			// header bytes
			fprintf( fp_log, "Header bytes\n" );

	
			fprintf( fp_log, "[%06X]  ", buf_ptr );
			for( int lcv_buf=0; lcv_buf<2; lcv_buf++ ) {
				fprintf( fp_log, "%02X ", UMEM8(BUF) );

				buf_ptr++;
			}
			fprintf( fp_log, "\n" );


			freq = UMEM16(BUF);
			fprintf( fp_log, "- Frequency = %d\n", freq );



			fprintf( fp_log, "- *** Digital PCM!!\n" );
			fprintf( fp_global, "[%06X]  *** Digital PCM = %d!!\n", buf_ptr-2, freq );

			
			if( track_gm == 0 && track_mt32 == 0 ) {
				fprintf( fp_global, "- MT-32 / GM not found!!! (@@@)\n" );
				global_notice++;
			}

			if( lcv == 0 ) {
				fprintf( fp_global, "- No MIDI alternative!!\n" );
				global_notice++;
			}

			if( pcm_panpot != 0x40 ) {
				fprintf( fp_global, "- MIDI panpot = %X!! (@@@)\n", pcm_panpot );
				global_notice++;
			}


			global_notice++;
			continue;
		}



		// header bytes
		fprintf( fp_log, "Header bytes\n" );
		fprintf( fp_log, "[%06X]  ", buf_ptr );

		for( int lcv_buf=0; lcv_buf<2; lcv_buf++ ) {
			fprintf( fp_log, "%02X ", UMEM8(BUF) );
			buf_ptr++;
		}

		fprintf( fp_log, "\n" );
		fprintf( fp_log, "\n" );
		fprintf( fp_log, "\n" );

		//------------------------------------------
		//------------------------------------------
		//------------------------------------------

		run_cmd = false;
		play_midi = true;

		while( play_midi ) {
			int start_ptr;


			// ABORT!! - some type of error
			if( buf_ptr >= buf_size ) {
				fprintf( fp_log, "ERROR - no EOF detected!\n" );
				fprintf( fp_global, "ERROR - no EOF detected!\n" );


				warnings++;
				global_notice++;

				break;
			}



			// special delay
			if( UMEM8(BUF) == 0xf8 ) {
				// header
				fprintf( fp_log, "[%06X]  ", buf_ptr );

				
				while( UMEM8(BUF) == 0xf8 ) {
					fprintf( fp_log, "{%02X} ", UMEM8(BUF) );

					buf_ptr++;
				}

				
				fprintf( fp_log, "\n" );
			}

			//------------------------------------------
			//------------------------------------------
			//------------------------------------------

			start_ptr = buf_ptr;


			// header
			fprintf( fp_log, "[%06X]  ", buf_ptr );


			// delay byte
			fprintf( fp_log, "{%02X} ", UMEM8(BUF) );

			buf_ptr++;



			// new MIDI command
			if( UMEM8(BUF) >= 0x80 ) {
				cmd = UMEM8(BUF);
				buf_ptr++;

				chan = cmd & 0xf;
				run_cmd = false;
			}
			else {
				fprintf( fp_log, " --  " );

				run_cmd = true;
			}


			arg1 = UMEM8(BUF);
			arg2 = UMEM8(BUF+1);


			switch( cmd >> 4 ) {
				case 0x8:
				case 0x9:
				case 0xa:
				case 0xb:
				case 0xe:
					if( (cmd >> 4) == 0xb && arg1 == 0xa && pcm_panpot == 0x40 ) {
						if( track_mt32 || track_gm )
							pcm_panpot = arg2;
					}


					buf_ptr += 2;


					if( run_cmd == false )
						fprintf( fp_log, "[%02X] %02X %02X\n", cmd, arg1, arg2 );
					else
						fprintf( fp_log, "%02X %02X\n", arg1, arg2 );
					break;


				case 0xc:
				case 0xd:
					buf_ptr += 1;


					if( run_cmd == false )
						fprintf( fp_log, "[%02X] %02X\n", cmd, arg1 );
					else
						fprintf( fp_log, "%02X\n", arg1 );
					break;


				case 0xf:
					switch( cmd ) {
						case 0xfc:
							fprintf( fp_log, "[%02X]  (STOP)\n", cmd );

							cmd = 0x00;
							play_midi = false;
							break;


						default:
							if( run_cmd == false )
								fprintf( fp_log, "[%02X]  (Unknown)\n", cmd );
							else
								fprintf( fp_log, "%02X\n", arg1 );
							break;
					}
					break;


				default:
					buf_ptr++;
					warnings++;
					global_notice++;


					if( run_cmd == false ) {
						fprintf( fp_log, "%02X = (Unknown)  (WARN!)\n", cmd );
						fprintf( fp_global, "%02X = (Unknown)  (WARN!)\n", cmd );
					}
					else
						fprintf( fp_log, "%02X  (WARN!)\n", arg1 );
						fprintf( fp_global, "%02X  (WARN!)\n", arg1 );
					break;
			}

			//------------------------------------------
			//------------------------------------------
			//------------------------------------------

			// look for soundtrack quirks
			// - NOTE: Sierra ch16 = special loop flags

			if( (track_gm || track_mt32 || track_adlib) && chan != 16-1 ) {
				if( (cmd >> 4) == 0x9 && arg2 > 0 ) {
					if( track_gm ) chan_gm[ chan ] = 1;
					if( track_mt32 ) chan_mt32[ chan ] = 1;
					if( track_adlib ) chan_adlib[ chan ] = 1;
				}


				if( (cmd >> 4) == 0xc ) {
					if( chan != 10-1 ) {
						if( track_gm ) patch_gm[ chan ][ arg1 ] = 1;
						if( track_mt32 ) patch_mt32[ chan ][ arg1 ] = 1;
						if( track_adlib ) patch_adlib[ chan ][ arg1 ] = 1;
					}
					else {
						if( track_gm ) perc_gm[ chan ][ arg1 ] = 1;
						if( track_mt32 ) perc_mt32[ chan ][ arg1 ] = 1;
						if( track_adlib ) perc_adlib[ chan ][ arg1 ] = 1;
					}
				}


				if( (cmd >> 4) == 0xd && arg1 > 0 ) {
					fprintf( fp_global, "[%06X]  Channel aftertouch = %02X %02X  (@@@)\n", start_ptr, cmd, arg1 );
					global_notice++;
				}


				if( (cmd >> 4) == 0xa && arg2 > 0 ) {
					fprintf( fp_global, "[%06X]  Polyphonic aftertouch = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );
					global_notice++;
				}


				if( cmd == 0xc9 && arg1 > 0 ) {
					fprintf( fp_global, "[%06X]  *** Drum kit = %02X %02X  (@@@)\n", start_ptr, cmd, arg1 );
					global_notice++;
				}


				if( cmd == 0x99 && arg2 > 0 ) {
					bool hit = false;
					bool hit2a = false;
					bool hit2b = false;
					bool hit3 = false;


					// notes = 0-127
					if( track_gm ) {
						if( arg1 >= 27 && arg1 <= 34 ) hit = true;
						if( arg1 >= 82 && arg1 <= 87 ) hit = true;
					}

					else if( track_mt32 ) {
						// CM32L + MT-32 custom
						if( arg1 >= 76 && arg1 <= 87 ) hit2a = true;
					}


					// CM32L only (outside of GM + GS)
					if( track_mt32 || track_gm ) {
						// CM32L only
						if( arg1 >= 88 && arg1 <= 108 ) hit2b = true;


						// Unknown area
						if( arg1 < 27 || arg1 > 108 ) hit3 = true;
					}


					if( hit ) fprintf( fp_global, "[%06X]  (GS) Percussion = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );
					if( hit2a ) fprintf( fp_global, "[%06X]  (CM32L / MT-32 Timbre) Percussion = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );
					if( hit2b ) fprintf( fp_global, "[%06X]  (CM32L) Percussion = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );
					if( hit3 ) fprintf( fp_global, "[%06X]  (UNK) Percussion = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );


					if( hit ) global_notice++;
					if( hit2a ) global_notice++;
					if( hit2b ) global_notice++;
					if( hit3 ) global_notice++;
				}


				if( (cmd>>4) == 0xb ) {
					bool hit, hit2;


					hit = true;
					hit2 = false;


					if( arg1 == 0x01 ) hit = false;
					if( arg1 == 0x07 ) hit = false;
					if( arg1 == 0x0a ) hit = false;
					if( arg1 == 0x40 ) hit = false;
					if( arg1 == 0x7b ) hit = false;


					// GM = no modulation
					if( arg1 == 0x01 && arg2 != 0x00 ) hit2 = true;


					if( hit && fp_global ) {
						bool hit_adlib = false;


						if( arg1 == 0x4b ) hit_adlib = true;
						if( arg1 == 0x4e ) hit_adlib = true;


						if( hit_adlib == false ) {
							fprintf( fp_global, "[%06X]  Control change = %02X %02X %02X  (*** WARN!!! ***)\n", start_ptr, cmd, arg1, arg2 );

							warnings++;
						}
						else
							fprintf( fp_global, "[%06X]  Control change = %02X %02X %02X  (Adlib)\n", start_ptr, cmd, arg1, arg2 );


						global_notice++;
					}

					else if( hit2 && fp_global )
						fprintf( fp_global, "[%06X]  Control change = %02X %02X %02X  (@@@)\n", start_ptr, cmd, arg1, arg2 );
				}
			}
		} // end play midi


		fprintf( fp_log, "\n" );
		fprintf( fp_log, "\n" );
		fprintf( fp_log, "\n" );
	}


	if( warnings ) fprintf( fp_log, "Warnings = %d", warnings );

	fclose(fp_log);

	//************************************************
	//************************************************
	//************************************************

	// look for soundtrack quirks


	bool header;


	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		if( chan_gm[lcv] == 1 && (lcv==1-1 || lcv>=11-1) ) {
			if( header == false ) {
				fprintf( fp_global, "*** GM = " );

				header = true;
			}


			fprintf( fp_global, "ch%d (+++) ", lcv+1 );
			global_notice++;
		}


		else if( chan_gm[lcv] == 1 && chan_mt32[lcv] == 0 ) {
			if( header == false ) {
				fprintf( fp_global, "*** GM = " );

				header = true;
			}


			fprintf( fp_global, "ch%d ", lcv+1 );
			global_notice++;
		}
	}


	if( header == true ) fprintf( fp_global, "\n" );

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		for( lcv2=0; lcv2<128; lcv2++ ) {
			if( patch_gm[lcv][lcv2] == 1 && patch_mt32[lcv][lcv2] == 0 ) {
				if( header == false ) {
					fprintf( fp_global, "*** GM patch (1-128)\n" );

					header = true;
				}


				if( map_patch_gm[lcv] != 255 )
					fprintf( fp_global, "- ch%d = [%02X] %d -> %d\n", lcv+1, lcv2, lcv2+1, map_patch_gm[lcv2]+1 );
				else
					fprintf( fp_global, "- ch%d = [%02X] %d -> %d  (unmapped!)\n", lcv+1, lcv2, lcv2+1, map_patch_gm[lcv2]+1 );

					
				global_notice++;
			}
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );



	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		for( lcv2=0; lcv2<128; lcv2++ ) {
			if( perc_gm[lcv][lcv2] == 1 && perc_mt32[lcv][lcv2] == 0 ) {
				if( header == false ) {
					fprintf( fp_global, "*** GM percussion (0-127)\n" );

					header = true;
				}


				if( map_perc_gm[lcv] != 255 )
					fprintf( fp_global, "- ch%d = [%02X] %d -> %d\n", lcv+1, lcv2, lcv2, map_perc_gm[lcv2]+1 );
				else
					fprintf( fp_global, "- ch%d = [%02X] %d -> %d  (unmapped!)\n", lcv+1, lcv2, lcv2, map_perc_gm[lcv2]+1 );


				global_notice++;
			}
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		if( chan_mt32[lcv] == 1 && (lcv==1-1 || lcv>=11-1) ) {
			if( header == false ) {
				fprintf( fp_global, "*** MT-32 = " );

				header = true;
			}


			fprintf( fp_global, "ch%d (@@@) ", lcv+1 );
			global_notice++;
		}


		else if( chan_mt32[lcv] == 1 && chan_gm[lcv] == 0 ) {
			if( header == false ) {
				fprintf( fp_global, "*** MT-32 = " );

				header = true;
			}


			fprintf( fp_global, "ch%d ", lcv+1 );
			global_notice++;
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		for( lcv2=0; lcv2<128; lcv2++ ) {
			if( patch_gm[lcv][lcv2] == 0 && patch_mt32[lcv][lcv2] == 1 ) {
				if( header == false ) {
					fprintf( fp_global, "*** MT-32 patch (1-128)\n" );

					header = true;
				}


				if( map_patch_mt32[lcv] != 255 )
					fprintf( fp_global, "- ch%d = [%02X] %d -> %d\n", lcv+1, lcv2, lcv2+1, map_patch_mt32[lcv2]+1 );
				else
					fprintf( fp_global, "- ch%d = [%02X] %d -> %d  (unmapped!)\n", lcv+1, lcv2, lcv2+1, map_patch_mt32[lcv2]+1 );


				global_notice++;
			}
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );



	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		for( lcv2=0; lcv2<128; lcv2++ ) {
			if( perc_gm[lcv][lcv2] == 0 && perc_mt32[lcv][lcv2] == 1 ) {
				if( header == false ) {
					fprintf( fp_global, "*** MT-32 percussion (0-127)\n" );

					header = true;
				}


				if( map_perc_mt32[lcv] != 255 )
					fprintf( fp_global, "- ch%d = [%02X] %d -> %d\n", lcv+1, lcv2, lcv2, map_perc_mt32[lcv2]+1 );
				else
					fprintf( fp_global, "- ch%d = [%02X] %d -> %d  (unmapped!)\n", lcv+1, lcv2, lcv2, map_perc_mt32[lcv2]+1 );


				global_notice++;
			}
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		if( chan_adlib[lcv] == 1 && chan_mt32[lcv] == 0 && chan_gm[lcv] == 0 ) {
			if( header == false ) {
				fprintf( fp_global, "*** Adlib = " );

				header = true;
			}


			fprintf( fp_global, "ch%d ", lcv+1 );
			global_notice++;
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		for( lcv2=0; lcv2<128; lcv2++ ) {
			if( patch_adlib[lcv][lcv2] == 1 && patch_gm[lcv][lcv2] == 0 && patch_mt32[lcv][lcv2] == 0 ) {
				if( header == false ) {
					fprintf( fp_global, "*** Adlib patch (1-128)\n" );

					header = true;
				}


				fprintf( fp_global, "- ch%d = [%02X] %d\n", lcv+1, lcv2, lcv2+1 );


				global_notice++;
			}
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );



	header = false;
	for( lcv=0; lcv<16; lcv++ ) {
		for( lcv2=0; lcv2<128; lcv2++ ) {
			if( perc_adlib[lcv][lcv2] == 1 && perc_gm[lcv][lcv2] == 0 && perc_mt32[lcv][lcv2] == 0 ) {
				if( header == false ) {
					fprintf( fp_global, "*** Adlib percussion (0-127)\n" );

					header = true;
				}


				fprintf( fp_global, "- ch%d = [%02X] %d\n", lcv+1, lcv2, lcv2 );


				global_notice++;
			}
		}
	}

	if( header == true ) fprintf( fp_global, "\n" );

	//------------------------------------------
	//------------------------------------------
	//------------------------------------------

	global_warnings += warnings;


	fprintf( fp_global, "\n" );
}
